# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Abstract base classes for writing posts.
'''


from firmant import paginate
from firmant import writers


class PostWriter(writers.Writer):
    '''Write/render individual post objects.

    This post writer is serves as the basis for other post writers that write
    posts individually.

    .. doctest::

       >>> class SamplePost(PostWriter):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj):
       ...         print 'Save post "%s %s" to "%s"' % \
                       (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
       >>> sp = SamplePost({}, {})
       >>> pprint(sp({}, {})) #doctest: +ELLIPSIS
       ({},
        {},
        [<firmant.writers.posts.SamplePost object at 0x...>,
         <firmant.writers.posts.SamplePost object at 0x...>])

    '''

    def key(self, post):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a post object:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            day
               The day of publication.

            slug
               The slug that is unique to the feed object.

        .. doctest::
           :hide:

           >>> class SamplePost(PostWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save post "%s %s" to "%s"' % \
                           (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
           >>> sp = SamplePost({}, {})

        .. doctest::

           >>> print objects.posts[0].published.date(), objects.posts[0].slug
           2009-12-31 party
           >>> pprint(sp.key(objects.posts[0]))
           {'day': 31, 'month': 12, 'slug': u'party', 'type': u'post', 'year': 2009}

        '''
        return {'type': u'post'
               ,'slug': post.slug
               ,'year': post.published.year
               ,'month': post.published.month
               ,'day': post.published.day
               }

    def obj_list(self, environment, objects):
        '''Return all objects stored under the key ``posts``.

        .. doctest::
           :hide:

           >>> class SamplePost(PostWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save post "%s %s" to "%s"' % \
                           (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
           >>> sp = SamplePost({}, {})

        .. doctest::

           >>> sp.obj_list(None, {})
           []
           >>> sp.obj_list(None, {'posts': []})
           []
           >>> sp.obj_list(None, {'posts': ['postobj']})
           ['postobj']

        '''
        return objects.get('posts', [])


class PostArchiveAll(writers.Writer):
    '''Write the posts into paginated pages.

    .. doctest::

       >>> class SamplePostArchiveAll(PostArchiveAll):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spaa = SamplePostArchiveAll({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived year page:

            type
               This is always ``post``.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveAll(PostArchiveAll):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spaa = SamplePostArchiveAll({}, {})

        .. doctest::

           >>> obj = spaa.obj_list({'settings': settings},
           ...                     {'posts': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([<firmant.parsers.RstObject object at 0x...>,
             <firmant.parsers.RstObject object at 0x...>],
            Paginated(None, 1, 2))

           >>> pprint(spaa.key(obj))
           {'page': 1, 'type': u'post'}

        '''
        return {'type': u'post'
               ,'page': obj[1].cur
               }

    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by year, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveAll(PostArchiveAll):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spaa = SamplePostArchiveAll({}, {})

        .. doctest::

           >>> spaa.obj_list({'settings': settings}, {})
           []
           >>> spaa.obj_list({'settings': settings}, {'posts': []})
           []
           >>> pprint(spaa.obj_list({'settings': settings},
           ...                      {'posts': objects.posts})) #doctest: +ELLIPSIS
           [([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(None, 1, 2)),
            ([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(1, 2, 3)),
            ([<firmant.parsers.RstObject object at 0x...>], Paginated(2, 3, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('posts', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.paginate(num_per_page, posts)


class PostArchiveYearly(writers.Writer):
    '''Write the posts into pages grouped by year and then paginated.

    .. doctest::

       >>> class SamplePostArchiveYearly(PostArchiveYearly):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spay = SamplePostArchiveYearly({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived year page:

            type
               This is always ``post``.

            year
               The year of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveYearly(PostArchiveYearly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spay = SamplePostArchiveYearly({}, {})

        .. doctest::

           >>> obj = spay.obj_list({'settings': settings},
           ...                     {'posts': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([<firmant.parsers.RstObject object at 0x...>],
            Paginated(None, (2010,), (2009,)),
            Paginated(None, 1, 2))
           >>> pprint(spay.key(obj))
           {'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'page': obj[2].cur
               }

    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by year, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveYearly(PostArchiveYearly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spay = SamplePostArchiveYearly({}, {})

        .. doctest::

           >>> spay.obj_list({'settings': settings}, {})
           []
           >>> spay.obj_list({'settings': settings}, {'posts': []})
           []
           >>> pprint(spay.obj_list({'settings': settings},
           ...                      {'posts': objects.posts})) #doctest: +ELLIPSIS
           [([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(None, (2010,), (2009,)),
             Paginated(None, 1, 2)),
            ([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(None, (2010,), (2009,)),
             Paginated(1, 2, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010,), (2009,), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('posts', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self.__splitfunc__, posts)

    @staticmethod
    def __splitfunc__(post):
        return (post.published.year,)


class PostArchiveMonthly(writers.Writer):
    '''Write the posts into pages grouped by month and then paginated.

    .. doctest::

       >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spam = SamplePostArchiveMonthly({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived month page:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spam = SamplePostArchiveMonthly({}, {})

        .. doctest::

           >>> obj = spam.obj_list({'settings': settings},
           ...                     {'posts': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([<firmant.parsers.RstObject object at 0x...>,
             <firmant.parsers.RstObject object at 0x...>],
            Paginated(None, (2010, 2), (2010, 1)),
            Paginated(None, 1, 2))
           >>> pprint(spam.key(obj))
           {'month': 2, 'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'month': obj[1].cur[1]
               ,'page': obj[2].cur
               }

    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by month, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spam = SamplePostArchiveMonthly({}, {})

        .. doctest::

           >>> spam.obj_list({'settings': settings}, {})
           []
           >>> spam.obj_list({'settings': settings}, {'posts': []})
           []
           >>> pprint(spam.obj_list({'settings': settings},
           ...                      {'posts': objects.posts})) #doctest: +ELLIPSIS
           [([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(None, (2010, 2), (2010, 1)),
             Paginated(None, 1, 2)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated(None, (2010, 2), (2010, 1)),
             Paginated(1, 2, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010, 2), (2010, 1), (2009, 12)),
             Paginated(None, 1, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010, 1), (2009, 12), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('posts', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self.__splitfunc__, posts)

    @staticmethod
    def __splitfunc__(post):
        return (post.published.year, post.published.month)


class PostArchiveDaily(writers.Writer):
    '''Write the posts into pages grouped by day and then paginated.

    .. doctest::

       >>> class SamplePostArchiveDaily(PostArchiveDaily):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spad = SamplePostArchiveDaily({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify an archived day page:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            day
               The day of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveDaily(PostArchiveDaily):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spad = SamplePostArchiveDaily({}, {})

        .. doctest::

           >>> obj = spad.obj_list({'settings': settings},
           ...                     {'posts': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([<firmant.parsers.RstObject object at 0x...>,
             <firmant.parsers.RstObject object at 0x...>],
            Paginated(None, (2010, 2, 2), (2010, 2, 1)),
            Paginated(None, 1, None))
           >>> pprint(spad.key(obj))
           {'day': 2, 'month': 2, 'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'month': obj[1].cur[1]
               ,'day': obj[1].cur[2]
               ,'page': obj[2].cur
               }

    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by day, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveDaily(PostArchiveDaily):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spad = SamplePostArchiveDaily({}, {})

        .. doctest::

           >>> spad.obj_list({'settings': settings}, {})
           []
           >>> spad.obj_list({'settings': settings}, {'posts': []})
           []
           >>> pprint(spad.obj_list({'settings': settings},
           ...                      {'posts': objects.posts})) #doctest: +ELLIPSIS
           [([<firmant.parsers.RstObject object at 0x...>,
              <firmant.parsers.RstObject object at 0x...>],
             Paginated(None, (2010, 2, 2), (2010, 2, 1)),
             Paginated(None, 1, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010, 2, 2), (2010, 2, 1), (2010, 1, 1)),
             Paginated(None, 1, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010, 2, 1), (2010, 1, 1), (2009, 12, 31)),
             Paginated(None, 1, None)),
            ([<firmant.parsers.RstObject object at 0x...>],
             Paginated((2010, 1, 1), (2009, 12, 31), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('posts', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self.__splitfunc__, posts)

    @staticmethod
    def __splitfunc__(post):
        return (post.published.year, post.published.month, post.published.day)


def _setup(test):
    '''Setup the environment for tests.
    '''
    from pysettings.settings import Settings
    from testdata.chunks import c900
    settings = Settings()
    settings.POSTS_PER_PAGE = 2
    test.globs['settings'] = settings
    test.globs['objects'] = c900



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Firmant is a framework for developing static web applications.

Much of today's web development focuses on developing dynamic applications that
regenerate the page for each view.  Firmant takes a different approach that
allows for publishing of static content that can be served by most http servers.

Some of the benefits of this approach include:

 * Build locally, deploy anywhere.  Many notable server distributions (including
   CentOS 5, and Debian Lenny) still ship old (pre-2.6) versions of Python.
   With Firmant, this is not an issue as static output may be published anywhere
   independent of the system where it was built.
 * Quicker page load times.  Search engines and viewers expect near-instant page
   load times and static content can meet these expectations.  Dynamic content
   can as well; however, it often requires more than simple hardware to do so.
 * Offline publishing capability.  Previewing changes to a website does not
   require Internet access, as the changes are all made locally.  Changes do not
   need to be pushed to a remote server.
 * Store content in revision control.  This is not strictly a feature granted by
   generating static pages.  Firmant is designed to make storing all content in
   a repository a trivial task -- something that web application frameworks that
   are powered by relational databases do not consider.

Modules in this package:

.. autosummary::
   :toctree: generated

   chunks
   du
   paginate
   routing
   utils
   writers

'''



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''The Firmant application.

Orchestrates the high process of parsing, transforming, and writing content.
'''


import logging
import datetime

from pysettings.modules import get_module

from firmant import utils
from firmant.chunks import AbstractChunk
from firmant.routing import URLMapper
from firmant.writers import j2
from firmant.utils import class_name


class Firmant(object):
    '''Perform a complete run from parsing through writing.

        >>> from pysettings.loaders import mod_to_settings
        >>> s = mod_to_settings('firmant.settings')
        >>> s.OUTPUT_DIR = outdir
        >>> f = Firmant(s)
        >>> f.log = Mock('log')
        >>> pprint(f.parsers) #doctest: +ELLIPSIS
        {'feeds': <firmant.parsers.feeds.FeedParser object at 0x...>,
         'posts': <firmant.parsers.posts.PostParser object at 0x...>,
         'static': <firmant.parsers.static.StaticParser object at 0x...>,
         'staticrst': <firmant.parsers.static.StaticRstParser object at 0x...>,
         'tags': <firmant.parsers.tags.TagParser object at 0x...>}
        >>> f.parse()
        >>> pprint(f.objs) #doctest: +ELLIPSIS
        {'feeds': [<firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>],
         'posts': [<firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>,
                   <firmant.parsers.RstObject object at 0x...>],
         'static': [static_obj<testdata/pristine/static/images/88x31.png>],
         'staticrst': [<firmant.parsers.RstObject object at 0x...>,
                       <firmant.parsers.RstObject object at 0x...>,
                       <firmant.parsers.RstObject object at 0x...>],
         'tags': [<firmant.parsers.RstObject object at 0x...>,
                  <firmant.parsers.RstObject object at 0x...>,
                  <firmant.parsers.RstObject object at 0x...>,
                  <firmant.parsers.RstObject object at 0x...>]}
        >>> f.cross_reference() #doctest: +ELLIPSIS
        >>> f.create_permalinks() #doctest: +ELLIPSIS
        >>> for post in f.objs['posts']:
        ...     pprint((post.tags, post.feeds)) #doctest: +ELLIPSIS
        ([<firmant.parsers.RstObject object at 0x...>],
         [<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>])
        ([<firmant.parsers.RstObject object at 0x...>],
         [<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>])
        ([<firmant.parsers.RstObject object at 0x...>],
         [<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>])
        ([<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>],
         [<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>])
        ([<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>],
         [<firmant.parsers.RstObject object at 0x...>,
          <firmant.parsers.RstObject object at 0x...>])
        >>> f()

    '''

    def __init__(self, settings):
        self.settings = settings
        self.urlmapper = URLMapper(settings.OUTPUT_DIR, settings.PERMALINK_ROOT,
                getattr(settings, 'URLS', None))
        self.log = logging.getLogger(class_name(self.__class__))

        self.objs = dict()
        self.writers = dict()
        self.chunks = list()
        self.env    = dict()
        self.env['log'] = self.log
        self.env['urlmapper'] = self.urlmapper
        self.env['settings'] = self.settings

        self.chunks.append(CheckURLConflicts())

        for glob in settings.GLOBALS:
            glob = utils.get_obj(glob)
            self.chunks.append(glob(self.env, self.objs))

        # Setup parsers
        self.parsers = dict()
        for key, parser in settings.PARSERS.items():
            parser = utils.get_obj(parser)
            self.parsers[key] = parser(self.settings)

        # Setup writers
        for writer in self.settings.WRITERS:
            writer = utils.get_obj(writer)
            self.chunks.append(writer(self.env, self.objs))

    def __call__(self):
        while len(self.chunks):
            self.chunks.sort(key = lambda x: x.scheduling_order)
            chunk = self.chunks[0]
            self.env, self.objs, newchunks = chunk(self.env, self.objs)
            self.chunks = self.chunks[1:]
            for c in newchunks:
                if c.scheduling_order <= chunk.scheduling_order:
                    error = _('New chunk violates scheduling_order constraint')
                    self.log.error(error)
                else:
                    self.chunks.append(c)

    def parse(self):
        '''Call parse on each configured parser.
        '''
        # Parse documents
        for key, parser in self.parsers.items():
            self.objs[key] = parser.parse()

    def cross_reference(self):
        '''Cross reference tags and feeds
        '''
        for feed in self.objs.get('feeds', []):
            feed.posts = list()
        for tag in self.objs.get('tags', []):
            tag.posts = list()
        # TODO inefficient, notproud
        for post in self.objs.get('posts', []):
            to_delete = list()
            for i, ptag in enumerate(post.tags):
                seen = False
                for tag in self.objs.get('tags', []):
                    if tag.slug == ptag:
                        seen = True
                        tag.posts.append(post)
                        post.tags[i] = tag
                if not seen:
                    warning  = _("Tag '%s' referenced but not defined.")
                    warning %= ptag
                    self.log.warning(warning)
                    to_delete.append(i)
            for i in reversed(to_delete):
                del post.tags[i]
            to_delete = list()
            for i, pfeed in enumerate(post.feeds):
                seen = False
                for feed in self.objs.get('feeds', []):
                    if feed.slug == pfeed:
                        seen = True
                        feed.posts.append(post)
                        post.feeds[i] = feed
                if not seen:
                    error  = _("Feed '%s' referenced but not defined.")
                    error %= pfeed
                    self.log.error(error)
                    to_delete.append(i)
            for i in reversed(to_delete):
                del post.feeds[i]

    def create_permalinks(self):
        '''Add permalinks to objects.
        '''
        # TODO:  Remove prior to 0.2.0
        url = self.urlmapper.url
        for post in self.objs.get('posts', []):
            post.permalink = url('atom', **{'type': 'post'
                                           ,'year': post.published.year
                                           ,'month': post.published.month
                                           ,'day': post.published.day
                                           ,'slug': post.slug
                                           })
        for tag in self.objs.get('tags', []):
            tag.permalink = url('html', **{'type': 'tag'
                                          ,'slug': tag.slug
                                          })
        for feed in self.objs.get('feeds', []):
            feed.permalink = url('atom', **{'type': 'feed'
                                           ,'slug': feed.slug
                                           })
        for staticrst in self.objs.get('staticrst', []):
            staticrst.permalink = url('html', **{'type': 'staticrst'
                                                ,'path': staticrst.path
                                                })


class CheckURLConflicts(AbstractChunk):
    '''A chunk that warns of writers with conflicting URLs.
    '''

    def __call__(self, environment, objects):
        '''Ensure that the sets of URLs are disjoint.
        '''
        urls = dict()
        for key, inst in environment['urls'].items():
            for url in inst:
                if url is None:
                    warning  = _("Writer %s's URLs incompletely defined.")
                    warning %= key
                    environment['log'].warning(warning)
                elif url in urls:
                    warning  = _('Writers %s and %s conflict over %s.')
                    warning %= urls[url], key, url
                    environment['log'].warning(warning)
                else:
                    urls[url] = key
        return (environment, objects, [])

    scheduling_order = 600


def _setup(self):
    '''Setup the test cases.

    Actions taken::

        - Create a temporary directory.

    '''
    import tempfile
    self.globs['outdir'] = tempfile.mkdtemp()


def _teardown(test):
    '''Cleanup the test cases.

    Actions taken::

        - Remove the temporary directory.

    '''
    import shutil
    shutil.rmtree(test.globs['outdir'])



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import stat
import tempfile
import os

from docutils import io
from docutils.core import publish_programmatically

from firmant.parsers import Parser
from firmant.parsers import RstObject
from firmant.parsers import RstParser
from firmant.du import MetaDataStandaloneReader
from firmant.utils import class_name
from firmant.utils.exceptions import log_uncaught_exceptions


__all__ = ['StaticParser', 'StaticRstParser']


class StaticObject(object):
    '''An object that serves as a placeholder for a file on the filesystem.
    '''

    __slots__ = ['fullpath', 'relpath', 'permalink']

    def __repr__(self):
        return 'static_obj<%s>' % self.fullpath


class StaticParser(Parser):
    '''Create stand-in objects for static files to be published.

    '''

    def paths(self):
        '''Return a list of paths to objects on the file system.
        '''
        settings = self.settings
        path  = os.path.join(settings.CONTENT_ROOT, settings.STATIC_SUBDIR)
        all_files = []
        for root, dirs, files in os.walk(path):
            for file in files:
                s = StaticObject()
                s.fullpath = os.path.join(root, file)
                if s.fullpath.startswith(path):
                    s.relpath = s.fullpath[len(path):]
                else:
                    raise RuntimeError('Funky paths in StaticParser')
                if s.relpath.startswith('/'):
                    s.relpath = s.relpath[1:]
                all_files.append(s)
        all_files.sort(key=lambda s: s.relpath)
        return all_files

    def parse_one(self, path):
        '''We use the identity function simply because we are copying files.
        '''
        return path


class StaticRstParser(RstParser):
    '''Create a static page from reStructured Text.

    For each reSt document, parse it to a page and extract its metadata.

        >>> from pysettings.settings import Settings
        >>> s = {'CONTENT_ROOT': 'content'
        ...     ,'STATIC_RST_SUBDIR': 'flat'
        ...     ,'REST_EXTENSION': 'rst'
        ...     }
        >>> s = Settings(s)
        >>> p = StaticRstParser(s)
        >>> p = p.parse_one('content/flat/links.rst')
        >>> p.copyright
        u'This document is part of the public domain.'
        >>> p.title
        u'Links'
        >>> p.content #doctest: +ELLIPSIS
        u'<ul class="simple">...</ul>\\n'

    Default values::

        >>> from pysettings.settings import Settings
        >>> s = {'CONTENT_ROOT': 'content'
        ...     ,'STATIC_RST_SUBDIR': 'flat'
        ...     ,'REST_EXTENSION': 'rst'
        ...     }
        >>> s = Settings(s)
        >>> p = StaticRstParser(s)
        >>> p = p.parse_one('content/flat/empty.rst')
        >>> p.title
        u''
        >>> p.content #doctest: +ELLIPSIS
        u''

    All static pages may be retrieved with::

        >>> from pysettings.settings import Settings
        >>> s = {'CONTENT_ROOT': 'content'
        ...     ,'STATIC_RST_SUBDIR': 'flat'
        ...     ,'REST_EXTENSION': 'rst'
        ...     }
        >>> s = Settings(s)
        >>> p = StaticRstParser(s)
        >>> p.log = Mock('log')
        >>> pprint(map(lambda p: p.title, p.parse()))
        Called log.info('parsed content/flat/about.rst')
        Called log.info('parsed content/flat/empty.rst')
        Called log.info('parsed content/flat/links.rst')
        [u'About', u'', u'Links']

    '''

    auto_metadata = [('copyright', 'copyright')]

    auto_pubparts = [('title', 'title')
                    ,('content', 'fragment')
                    ]

    def paths(self):
        '''Return a list of paths to tag objects on the filesystem.

        Consider all the contents of the tags directory (by default this is
        ``content_root/flat``).  Only files ending in ``suffix`` are considered.

        Directory entries that are not files are ignored.

            >>> from pysettings.settings import Settings
            >>> s = {'CONTENT_ROOT': 'content'
            ...     ,'STATIC_RST_SUBDIR': 'flat'
            ...     ,'REST_EXTENSION': 'rst'
            ...     }
            >>> s = Settings(s)
            >>> p = StaticRstParser(s)
            >>> p.paths()
            ['content/flat/about.rst', 'content/flat/empty.rst', 'content/flat/links.rst']

        '''
        settings = self.settings
        path = os.path.join(settings.CONTENT_ROOT, settings.STATIC_RST_SUBDIR)
        all_files = []
        for root, dirs, files in os.walk(path):
            files = filter(lambda f: f.endswith(settings.REST_EXTENSION), files)
            files = map(lambda f: os.path.join(root, f), files)
            files = filter(lambda f: os.path.isfile(f), files)
            all_files += files
        all_files.sort()
        return all_files

    def new_object(self, path, d, pub):
        '''Return an instance of the object to which rst documents are parsed.
        '''
        settings = self.settings
        root = os.path.join(settings.CONTENT_ROOT, settings.STATIC_RST_SUBDIR)
        if path.startswith(root):
            path = path[len(root):]
        if path.startswith('/'):
            path = path[1:]
        p = RstObject()
        p.path = unicode(path.rsplit('.', 1)[0])
        return p

    def default(self, attr):
        '''Return the default value of an attribute.
        '''
        d = {}
        return d.get(attr, u'')
